/*== BU.java ======================================================
BU implements a number of often used byte-buffer utilities.
Version     : $Id: BU.java 34 2011-03-31 14:34:15Z hartwigthomas $
Application : Utilities
Description : Implements a number of often used byte-buffer utilities.
------------------------------------------------------------------------
Copyright  : Enter AG, Zurich, Switzerland, 2008
Created    : 12.03.2008, Hartwig Thomas
------------------------------------------------------------------------
The class ch.enterag.utils.BU is free software; you can redistribute it
and/or modify it under the terms of the GNU General Public License 
version 2 or later as published by the Free Software Foundation.

ch.enterag.utils.BU is distributed in the hope that it will 
be useful, but WITHOUT ANY WARRANTY; without even the implied warranty 
of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

If you have a need for licensing ch.enterag.utils.BU without 
some of the restrictions specified in the GNU General Public License,
it is possible to negotiate a different license with the copyright holder.
======================================================================*/

package ch.enterag.utils;

/*====================================================================*/
/** Implements a number of often used byte-buffer utilities.
 @author Hartwig Thomas
 */
public class BU
{
  /*====================================================================
  little-endian utilities
  ====================================================================*/
  /** byte mask for extracting low order byte from long */
	private static final long lBYTE_MASK = 0x00000000000000FFL;
  /** byte mask for extracting low order byte from int */
	private static final int iBYTE_MASK = 0x000000FF;
  /** byte mask for extracting low order byte from 2-byte integer */
	private static final int iNYBBLE_MASK = 0x0000000F;

  /*------------------------------------------------------------------*/
  /** converts int to byte (unsigned)
   * @param i int to be converted.
   * @return (unsigned) byte.
   */
	private static byte lowByte(int i)
	{
		i = i & iBYTE_MASK;
		if (i >= 0x00000080)
			i = - 0x00000100 + i;
		return (byte)i;
	} /* lowByte */
	
  /*------------------------------------------------------------------*/
  /** extracts an short from 2 bytes of a byte buffer in little-endian
   * order.
   @param bufShort little-endian byte buffer of size => 2.
   @param iPos position of short
   @return extracted short.
   */
  public static short toShort(byte[] bufShort, int iPos)
  {
  	short wShort = 0;
  	int iByte = 0;
  	if (bufShort.length < iPos + 2)
  		throw new IllegalArgumentException("byte buffer for short must have length >= 2!");
  	for (int i = iPos+1; i >= iPos; i--)
  	{
  		/* shift output */
  		wShort <<= 8;
  		iByte = bufShort[i];
  		/* bytes are signed! */
  		iByte = iByte & iBYTE_MASK;
  		/* set lowest byte */
  		wShort = (short)(wShort | iByte);
  	}
  	return wShort;
  } /* toShort */
  
  /*------------------------------------------------------------------*/
  /** extracts an short from first 2 bytes of a byte buffer in little-endian
   * order.
   @param bufShort little-endian byte buffer of size => 2.
   @return extracted short.
   */
  public static short toShort(byte[] bufShort)
  {
  	return toShort(bufShort,0);
  } /* toShort */
  
  /*------------------------------------------------------------------*/
  /** converts a short into a 2-byte little-endian byte buffer. 
   @param wShort short to be converted.
   @return little-endian byte buffer.
   */
  public static byte[] fromShort(short wShort)
  {
  	byte[] bufShort = new byte[2];
  	for (int i = 1; i >= 0; i--)
  	{
  		bufShort[1-i] = lowByte(wShort);
  		/* shift input */
  		wShort >>= 8;
  	}
  	return bufShort;
  } /* fromShort */
  
  /*------------------------------------------------------------------*/
  /** extracts an int from 4 bytes of a byte buffer in little-endian
   * order.
   @param bufInt little-endian byte buffer of size => 4.
   @param iPos starting position.
   @return extracted int.
   */
  public static int toInt(byte[] bufInt, int iPos)
  {
  	int iInt = 0;
  	int iByte = 0;
  	if (bufInt.length < iPos + 4)
  		throw new IllegalArgumentException("byte buffer for int must have length >= 4!");
  	for (int i = iPos+3; i >= iPos; i--)
  	{
  		/* shift output */
  		iInt <<= 8;
  		iByte = bufInt[i];
  		/* bytes are signed! */
  		iByte = iByte & iBYTE_MASK;
  		/* set lowest byte */
  		iInt |= iByte;
  	}
  	return iInt;
  } /* toInt */
  
  /*------------------------------------------------------------------*/
  /** extracts an int from first 4 bytes of a byte buffer in little-endian
   * order.
   @param bufInt little-endian byte buffer of size => 4.
   @return extracted int.
   */
  public static int toInt(byte[] bufInt)
  {
  	return toInt(bufInt,0);
  } /* toInt */
  
  /*------------------------------------------------------------------*/
  /** converts an int into a 4-byte little-endian byte buffer. 
   @param iInt int to be converted.
   @return little-endian byte buffer.
   */
  public static byte[] fromInt(int iInt)
  {
  	byte[] bufInt = new byte[4];
  	for (int i = 3; i >= 0; i--)
  	{
  		bufInt[3-i] = lowByte(iInt);
  		/* shift input */
  		iInt >>= 8;
  	}
  	return bufInt;
  } /* fromInt */
  
  /*------------------------------------------------------------------*/
  /** extracts a long from 8 bytes of a byte buffer in little-endian
   * order.
   @param bufLong little-endian byte buffer of size => 8.
   @param iPos starting position.
   @return extracted long.
   */
  public static long toLong(byte[] bufLong, int iPos)
  {
  	long lLong = 0;
  	long lByte = 0;
  	if (bufLong.length < iPos + 8)
  		throw new IllegalArgumentException("byte buffer for long must have length >= 8!");
  	for (int i = iPos+7; i >= iPos; i--)
  	{
  		/* shift output */
  		lLong <<= 8;
  		lByte = bufLong[i];
  		/* bytes are signed! */
  		lByte = lByte & iBYTE_MASK;
  		/* set lowest byte */
  		lLong |= lByte;
  	}
  	return lLong;
  } /* toLong */
  
  /*------------------------------------------------------------------*/
  /** extracts a long from first 8 bytes of a byte buffer in little-endian
   * order.
   @param bufLong little-endian byte buffer of size => 8.
   @return extracted long.
   */
  public static long toLong(byte[] bufLong)
  {
  	return toLong(bufLong,0);
  } /* toLong */
  
  /*------------------------------------------------------------------*/
  /** converts a long into a 8-byte little-endian byte buffer. 
   @param lLong long to be converted.
   @return little-endian byte buffer.
   */
  public static byte[] fromLong(long lLong)
  {
  	byte[] bufLong = new byte[8];
  	for (int i = 7; i >= 0; i--)
  	{
  		bufLong[7-i] = lowByte((int)(lLong & lBYTE_MASK));
  		/* shift input */
  		lLong >>= 8;
  	}
  	return bufLong;
  } /* fromLong */
	
  /*====================================================================
  hex utilities
  ====================================================================*/
  /** hex digits */
  private static String sHexDigits = "0123456789ABCDEF";

  /*------------------------------------------------------------------*/
  /** returns the two hex digits representing the given byte.
  @param b byte to be converted.
  @return hex string of length 2.
  */
  public static String toHex(byte b)
  {
  	StringBuffer sbHex = new StringBuffer();
  	int i = b & iBYTE_MASK;
  	int iLow = i & iNYBBLE_MASK;
  	int iHigh = i >> 4;
    sbHex.append(sHexDigits.charAt(iHigh));
    sbHex.append(sHexDigits.charAt(iLow));
  	return sbHex.toString();
  } /* toHex */
  
  /*------------------------------------------------------------------*/
  /** extracts a hex string from the byte buffer.
   @param buffer buffer to be converted to hex string.
   @param iOffset start offset (included).
   @param iLength number of bytes to be converted.
   @return hex string of length 2*iLength.
   */
  public static String toHex(byte[] buffer, int iOffset, int iLength)
  {
  	StringBuffer sbHex = new StringBuffer();
  	for (int i = iOffset; i < iOffset + iLength; i++)
  		sbHex.append(toHex(buffer[i]));
  	return sbHex.toString();
  } /* toHex */
  
  /*------------------------------------------------------------------*/
  /** extracts a hex string from the byte buffer.
   @param buffer buffer to be converted to hex string.
   @return hex string of length 2*iLength.
   */
  public static String toHex(byte[] buffer)
  {
  	return toHex(buffer,0,buffer.length);
  } /* toHex */
  
  /*------------------------------------------------------------------*/
  /** converts two hex characters into a byte. 
   @param cHigh high hex digit.
   @param cLow high hex digit.
   @return byte.
   */
  public static byte fromHex(char cHigh, char cLow)
  {
  	byte b = 0;
  	int iHigh = sHexDigits.indexOf(cHigh);  
  	int iLow = sHexDigits.indexOf(cLow);
  	if ((iLow >= 0) && (iHigh >= 0))
  	{
  		int i = iHigh << 4;
  		i = i + iLow;
  		b = lowByte(i);
  	}
  	else
  		throw new IllegalArgumentException("Invalid hex data " + Character.toString(cHigh) + Character.toString(cLow) + "!");
  	return b;
  } /* fromHex */
  
  /*------------------------------------------------------------------*/
  /** converts a hex string into a byte buffer. 
   @param sHex hex string (even length).
   @return byte buffer.
   */
  public static byte[] fromHex(String sHex)
  {
  	if ((sHex.length() & 1) != 0)
  		throw new IllegalArgumentException("Hex string must have even number of hex digits!");
  	sHex = sHex.toUpperCase();
  	byte[] buffer = new byte[sHex.length()/2];
  	for (int i = 0; i < buffer.length; i++)
  		buffer[i] = fromHex(sHex.charAt(2*i),sHex.charAt(2*i+1));
  	return buffer;
  } /* fromHex */
  
} /* BU */
