/*== SU.java ===========================================================
SU implements a number of often used string utilities.
Version     : $Id: SU.java 34 2011-03-31 14:34:15Z hartwigthomas $
Application : Utilities
Description : SU implements a number of often used string utilities.
------------------------------------------------------------------------
Copyright  : Enter AG, Zurich, Switzerland, 2008
Created    : 15.02.2008, Hartwig Thomas
------------------------------------------------------------------------
The class ch.enterag.utils.BU is free software; you can redistribute it
and/or modify it under the terms of the GNU General Public License 
version 2 or later as published by the Free Software Foundation.

ch.enterag.utils.BU is distributed in the hope that it will 
be useful, but WITHOUT ANY WARRANTY; without even the implied warranty 
of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

If you have a need for licensing ch.enterag.utils.BU without 
some of the restrictions specified in the GNU General Public License,
it is possible to negotiate a different license with the copyright holder.
======================================================================*/

package ch.enterag.utils;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.io.UnsupportedEncodingException;
import java.text.MessageFormat;


/*====================================================================*/
/** SU implements a number of often used string utilities.
 @author Hartwig Thomas
 */
public class SU
{
  /*====================================================================
  (static, public) Methods
  ====================================================================*/
  /*------------------------------------------------------------------*/
  /** returns true, if s neither has length 0 nor is null.
   @param s string to be tested.
   @return true, if string is not null and has length greater than zero. 
   */
  public static boolean isNotEmpty(String s)
  {
	  return (s != null) && (s.length() > 0);
  } /* isNotEmpty */
  
  /*------------------------------------------------------------------*/
  /** returns true, if s neither is white nor is null.
   @param s string to be tested.
   @return true, if string is not null and has length greater than zero. 
   */
  public static boolean isNotWhite(String s)
  {
	  return (s != null) && (s.trim().length() > 0);
  } /* isNotWhite */
  
  /*------------------------------------------------------------------*/
  /** replaces all occurrences of sFind by sReplace.
   @param s string to be transformed.
   @param sFind substring to be replaced.
   @param sReplace substring to insert instead of sFind.
   @return transformed string.
   @deprecated is available as a String method since JAVA 1.5. 
   */
  @Deprecated
  public static String replace(String s, String sFind, String sReplace)
  {
  	StringBuffer sb = new StringBuffer();
  	if (SU.isNotEmpty(s))
  	{
    	int iStart = 0;
    	for (int iMatch = s.indexOf(sFind, iStart); iMatch >= 0; iMatch = s.indexOf(sFind, iStart))
    	{
    		/* append next portion of s */
      	sb.append(s.substring(iStart, iMatch));
    		/* append sReplace */
      	sb.append(sReplace);
    		/* skip sFind */
    		iStart = iMatch + sFind.length();
    	}
    	/* append last portion */
    	sb.append(s.substring(iStart, s.length()));
  	}
  	return sb.toString();
  } /* replace */
  
  /*------------------------------------------------------------------*/
  /** format works like MessageFormat.format but assumes, that all
   * single quotes not immediately preceding a { or } are meant to represent
   * a single quote.
   @param sPattern string to be formatted.
   @param aoReplace replacement arguments to be used.
   @return formatted string.
   */
  public static String format(String sPattern, Object[] aoReplace)
  {
  	StringBuffer sb = new StringBuffer();
  	for (int i = 0; i < sPattern.length(); i++)
  	{
  		char c = sPattern.charAt(i);
			sb.append(c);
  		if (sPattern.charAt(i) == '\'')
  		{
  			/* if it is not the last character and not followed by { or } then duplicate it */
  			if (i < sPattern.length()-1)
  			{
  				c = sPattern.charAt(i+1);
  				if ((c != '{') && (c != '}'))
  				sb.append('\'');
  			}
  		}
  	}
  	return MessageFormat.format(sb.toString(),aoReplace);
  } /* format */
  
  /*------------------------------------------------------------------*/
  /** format replaces {0} in the pattern by the given string.
   @param sPattern string to be formatted.
   @param sReplace replacement string to be used.
   @return formatted string.
   */
  public static String format(String sPattern, String sReplace)
  {
  	return format(sPattern,new Object[]{sReplace});
  } /* format */
  
  /*------------------------------------------------------------------*/
  /** returns a string of the given length filled with a given character.
   @param iRepetitions length of returned String.
   @param c character to be used.
   @return String of given length filled with desired character. 
   */
  public static String repeat(int iRepetitions, char c)
  {
	  StringBuffer sb = new StringBuffer(iRepetitions);
	  for (int i = 0; i < iRepetitions; i++)
	  	sb.append(c);
	  return sb.toString();
  } /* repeat */
  
  /*------------------------------------------------------------------*/
  /** returns a string of the given number of repetitions of a base string.
   @param iRepetitions length of returned String.
   @param s character to be used.
   @return String of given length filled with desired character. 
   */
  public static String repeat(int iRepetitions, String s)
  {
	  StringBuffer sb = new StringBuffer(iRepetitions);
	  for (int i = 0; i < iRepetitions; i++)
	  	sb.append(s);
	  return sb.toString();
  } /* repeat */
  
  /*------------------------------------------------------------------*/
  /** returns the next possible line break point in the given range.
   @param s String to be broken.
   @param iStart beginning of the line.
   @param iEnd maximum value of breakpoint.
   @return breakpoint.
   */
  public static int getBreakPoint(String s, int iStart, int iEnd)
  {
  	if (iEnd > s.length())
  		iEnd = s.length();
  	/* if there are line breaks between iStart and iEnd then take the first */
  	int iBreak = s.indexOf('\n',iStart)+1;
  	if ((iBreak <= iStart) || (iBreak > iEnd))
  	{
    	iBreak = iEnd;
    	if ((iEnd < s.length()) && (s.charAt(iEnd) != ' '))
    	{
    		for (int i = iEnd-1; (iBreak == iEnd) && (i > iStart); i--)
    		{
    			char c = s.charAt(i+1);
    			if ((c == ' ') || (c == '-') || (c == '\n') || (c == '\r') || (c == '\t'))
    				iBreak = i+1;
    		}
    	}
  	}
  	return iBreak;
  } /* getBreakPoint */
  
  /*====================================================================
  Utilities
  ====================================================================*/
  /*------------------------------------------------------------------*/
  /** converts a portion of a character array to an encoded byte buffer.
   @param ac character array to be converted.
   @param iOffset offset of portion in character array.
   @param iLength length of portion in character array.
   @param sEncoding encoding to be used.
   @return encoded byte buffer.
   */
  private static byte[] putEncodedCharArray(char[] ac, int iOffset, int iLength, String sEncoding)
	{
  	byte[] buf = null;
  	ByteArrayOutputStream baos = new ByteArrayOutputStream(iLength);
  	try
  	{
  		OutputStreamWriter osw = new OutputStreamWriter(baos,sEncoding);
  		for (int i = iOffset; i < iOffset + iLength; i++)
  		  osw.write(ac[i]);
  		osw.close();
  		buf = baos.toByteArray();
  	}
  	catch (UnsupportedEncodingException uee) { System.out.println(uee.getClass().getName()+":"+uee.getMessage());}
  	catch (IOException ie) { System.out.println(ie.getClass().getName()+":"+ie.getMessage());}
  	return buf;
	} /* putEncodedCharArray */
  
  /*------------------------------------------------------------------*/
  /** converts a character array to an UTF8-encoded byte buffer.
   @param ac character array to be converted.
   @param iOffset offset in character array where encoding is to start.
   @param iLength length of character array to be converted.
   @return encoded byte buffer.
   */
  public static byte[] putUtf8CharArray(char[] ac, int iOffset, int iLength)
	{
  	return putEncodedCharArray(ac,iOffset,iLength,"UTF-8");
	} /* putUtf8CharacterArray */
  
  /*------------------------------------------------------------------*/
  /** converts a string to an encoded byte buffer.
   @param s string to be converted.
   @param sEncoding encoding to be used.
   @return encoded byte buffer.
   */
  public static byte[] putEncodedString(String s, String sEncoding)
	{
  	byte[] buf = null;
  	ByteArrayOutputStream baos = new ByteArrayOutputStream(s.length());
  	try
  	{
  		OutputStreamWriter osw = new OutputStreamWriter(baos,sEncoding);
  		osw.write(s);
  		osw.close();
  		buf = baos.toByteArray();
  	}
  	catch (UnsupportedEncodingException uee) { System.out.println(uee.getClass().getName()+":"+uee.getMessage());}
  	catch (IOException ie) { System.out.println(ie.getClass().getName()+":"+ie.getMessage());}
  	return buf;
	} /* putEncodedString */
  
  /*------------------------------------------------------------------*/
  /** converts an string to an UTF8-encoded byte buffer.
   @param s string.
   @return encoded byte buffer.
   */
  public static byte[] putUtf8String(String s)
	{
  	return putEncodedString(s,"UTF-8");
	} /* putUtf8String */
  
  /*------------------------------------------------------------------*/
  /** converts a string to an Cp437-encoded byte buffer.
   @param s string.
   @return encoded byte buffer.
   */
  public static byte[] putCp437String(String s)
	{
  	return putEncodedString(s,"Cp437");
	} /* putCp437String */
  
  /*------------------------------------------------------------------*/
  /** converts an encoded byte buffer to a string.
   @param buf byte buffer.
   @param sEncoding encoding.
   @return encoded string.
   */
  public static String getEncodedString(byte[] buf, String sEncoding)
	{
  	StringBuffer sb = new StringBuffer();
  	ByteArrayInputStream bais = new ByteArrayInputStream(buf);
  	try
  	{
  	  InputStreamReader isr = new InputStreamReader(bais,sEncoding);
  	  for (int i = isr.read(); i != -1; i = isr.read())
  	  	sb.append((char)i);
  	}
  	catch (UnsupportedEncodingException uee) { System.out.println(uee.getClass().getName()+":"+uee.getMessage());}
  	catch (IOException ie) { System.out.println(ie.getClass().getName()+":"+ie.getMessage());}
  	return sb.toString();
	} /* getEncodedString */
  
  /*------------------------------------------------------------------*/
  /** converts an UTF8-encoded byte buffer to a string.
   @param buf byte buffer.
   @return UTF8-encoded string.
   */
  public static String getUtf8String(byte[] buf)
	{
  	return getEncodedString(buf,"UTF-8");
	} /* getUtf8String */
  
  /*------------------------------------------------------------------*/
  /** converts an Cp437-encoded byte buffer to a string.
   @param buf byte buffer.
   @return Cp437-encoded string.
   */
  public static String getCp437String(byte[] buf)
	{
  	return getEncodedString(buf,"Cp437");
	} /* getCp437String */
  
	/*------------------------------------------------------------------*/
  /** prepare string for a CSV cell by replacing all special characters
   * by escaping their hex value using the \x prefix.
   * @param sText string to be prepared.
   * @return prepared string
   */
  public static String toCsv(String sText)
  {
  	/* code \ as \\ */
		sText = SU.replace(sText, "\\", "\\\\");
		sText = SU.replace(sText, "\n", "\\n");
		sText = SU.replace(sText, "\r", "\\r");
		sText = SU.replace(sText, "\t", "\\t");
  	/* code \x00 - \x1F */
		for (int i = 0; i < 32; i++)
		{
			char c = (char)i;
			String s = String.valueOf(c);
			String sEntity = "\\x"+BU.toHex((byte)i);
			sText = SU.replace(sText, s, sEntity);
		}
  	return sText;
  } /* toCsv */
  
	/*------------------------------------------------------------------*/
  /** prepare text for xs:string by replacing all "critical" but 
   * permitted characters by entity references.
   * @param sText text to be prepared.
   * @param bDom true, if DOM expansion will follow.
   * @return prepared text.
   */
  public static String toXml(String sText, boolean bDom)
  {
  	if (isNotEmpty(sText))
  	{
  		if (!bDom)
  		{
				sText = SU.replace(sText, "&", "&amp;");
				sText = SU.replace(sText, "<", "&lt;");
				sText = SU.replace(sText, "\"", "&quot;");
  		}
			/* we use \\u00HH for exotic non-characters */
			sText = SU.replace(sText, "\\", "\\u005C");
  		/* replace all characters 0 - 31 by \\u00HH */
  		for (int i = 0; i < 32; i++)
  		{
  			char c = (char)i;
  			String s = String.valueOf(c);
  			String sEntity = "\\u00"+BU.toHex((byte)i);
  			sText = SU.replace(sText, s, sEntity);
  		}
  		/* replace all characters 127-159 by \\u00HH */
  		for (int i = 127; i < 160; i++)
  		{
  			char c = (char)i;
  			String s = String.valueOf(c);
  			byte b = (byte)(0xFFFFFF00 | i);
  			if (i < 128)
  				b = (byte)i;
  			String sEntity = "\\u00"+BU.toHex(b);
  			sText = SU.replace(sText, s, sEntity);
  		}
  		/* replace multiple spaces by \\u0020 */
			for (String sText1 = SU.replace(sText, "  ", " \\u0020");
			     !sText.equals(sText1);
			     sText1 = SU.replace(sText, "  ", " \\u0020"))
				sText = sText1;
			/* replace leading and trailing spaces by \\u0020 */
			if (sText.charAt(0) == ' ')
				sText = "\\u0020" + sText.substring(1);
			if (sText.charAt(sText.length()-1) == ' ')
				sText = sText.substring(0, sText.length()-1) + "\\u0020";
  	}
    return sText;  	
  } /* toXml */
  
	/*------------------------------------------------------------------*/
  /** prepare text from xs:string by removing all "ignorable" white space 
   * and resolving all entity references.
   * @param sText text to be prepared.
   * @param bDom true, of DOM expansion has preceded.
   * @return prepared text.
   */
  public static String fromXml(String sText,boolean bDom)
  {
  	if (isNotEmpty(sText))
  	{
  		/* replace all "white" characters by spaces */
    	sText = SU.replace(sText,"\n"," ");
    	sText = SU.replace(sText,"\r"," ");
    	sText = SU.replace(sText,"\t"," ");
    	/* replace multiple spaces by single spaces */
  		for (String sText1 = SU.replace(sText, "  ", " ");
  		     !sText.equals(sText1);
  		     sText1 = SU.replace(sText, "  ", " "))
  			sText = sText1;
  		/* resolve entity references (including blanks) */
  		for (int i = 0; i <= 32; i++)
  		{
  			char c = (char)i;
  			String s = String.valueOf(c);
  			String sEntity = "\\u00"+BU.toHex((byte)i);
  			sText = SU.replace(sText, sEntity, s);
  		}
  		/* replace all characters 127-159 by &#xHH; */
  		for (int i = 127; i < 160; i++)
  		{
  			char c = (char)i;
  			String s = String.valueOf(c);
  			byte b = (byte)(0xFFFFFF00 | i);
  			if (i < 128)
  				b = (byte)i;
  			String sEntity = "\\u00"+BU.toHex(b);
  			sText = SU.replace(sText, sEntity, s);
  		}
  		sText = SU.replace(sText,"\\u005C","\\");
  		if (!bDom)
  		{
	  		sText = SU.replace(sText, "&quot;", "\"");
	  		sText = SU.replace(sText, "&lt;", "<");
	  		sText = SU.replace(sText, "&amp;", "&");
  		}
  	}
    return sText;  	
  } /* fromXml */
  
} /* SU */
